from bs4 import BeautifulSoup, Tag
import copy

from page_scraper.core.utils import exclude_tags, process_text, canonical_url, process_link, classify_url, clear_text
from page_scraper.entities.models import PageHeading, UrlContext
from page_scraper.infrastructure.fetcher import HttpFetcher

TAGS_TO_EXCLUDE = ("[class*=comment], [id*=comment], [id*=sidebar], [id*=footer],"
                          "nav, footer, aside, menu, sidebar, head, script")
HEADINGS = "h1,h2,h3,h4,h5,h6"

fetcher = HttpFetcher()

VIDEO_PROVIDERS = (
    "youtube.com",
    "youtu.be",
    "vimeo.com",
    "aparat.com",
    "dailymotion.com",
)

AUDIO_PROVIDERS = (
    "spotify.com",
    "soundcloud.com",
    "podcasts.google.com",
    "anchor.fm",
)

def get_headings(soup: BeautifulSoup) -> list[Tag]:
    cleaned = copy.copy(soup)
    exclude_tags(cleaned, TAGS_TO_EXCLUDE)
    return [
        h for h in cleaned.select(HEADINGS)
    ]

def get_clean_tags(soup: BeautifulSoup, tag_string:str) -> list[Tag]:
    cleaned = copy.copy(soup)
    exclude_tags(cleaned, TAGS_TO_EXCLUDE)
    return [
        i for i in cleaned.select(tag_string)
    ]

def get_media(soup: BeautifulSoup, media_tag:str) -> list[Tag]:
    cleaned = copy.copy(soup)
    exclude_tags(cleaned, TAGS_TO_EXCLUDE)
    return [
        i for i in cleaned.select(media_tag)
    ]

def create_tag_dict(tag:Tag) -> dict:
    tag_name = tag.name
    tag_text = process_text(tag)
    return {"name": tag_name, "text": tag_text}

def process_headings(tags:list[Tag]) -> list[PageHeading]:
    result = []
    for tag in tags:
        tag_dict = create_tag_dict(tag)
        page_heading = PageHeading(tag_dict.get("name"),tag_dict.get("text"))
        result.append(page_heading)
    return result

def process_alt(tag:Tag)->str | None:
    return clear_text(tag.get("alt")) if tag.get("alt") else None

def get_media_src(tag:Tag) -> str|None:
    return tag.get("src",None)

def process_media_tag_with_url(url:str, base:str) -> UrlContext:
    address = canonical_url(url, base)
    classified_url = classify_url(address)
    if classified_url == "web":
        response = fetcher.head(address)
        return process_link(url, response=response)
    else:
        return process_link(url, url_type=classified_url)


def process_image(tags:list[Tag], base:str)->list[UrlContext]:
    result = []
    for tag in tags:
        src = get_media_src(tag)
        if not src:
            continue

        data = process_media_tag_with_url(src, base)

        if not data:
            continue

        data.text = process_alt(tag)

        data.position = "IMAGE_TAG"

        result.append(data)

    return result

def process_media_tags(tags:list[Tag], base:str, tag_type:str) -> list[UrlContext]:
    result = []

    for tag in tags:
        if get_media_src(tag):

            data = process_media_tag_with_url(get_media_src(tag), base)

            if not data:
                continue

            data.position = tag_type

            result.append(data)

            continue

        for source in tag.find_all("source"):
            if get_media_src(source):
                data = process_media_tag_with_url(get_media_src(source), base)

                if not data:
                    continue

                data.position = tag_type

                result.append(data)

    return result

def process_video_iframe(tags:list[Tag], base:str) -> list[UrlContext]:
    result = []
    for tag in tags:
        src = get_media_src(tag)
        if not src:
            continue

        if not any(provider in src for provider in VIDEO_PROVIDERS):
            continue

        data = process_media_tag_with_url(src, base)

        if not data:
            continue

        data.position = "VIDEO_IFRAME"

        result.append(data)

    return result

def process_video_embed(tags:list[Tag], base:str) -> list[UrlContext]:
    result = []
    for tag in tags:
        src = get_media_src(tag) or tag.get("data")
        if not src:
            continue

        data = process_media_tag_with_url(src, base)

        if not data:
            continue

        data.position = "VIDEO_EMBED"

        result.append(data)

    return result


import copy

from page_scraper.entities.models import PageContext
from page_scraper.meta.utils import process_title_tag, process_description_tag


class TitleScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        title = process_title_tag(soup_copy)
        if title:
            page.meta_title = title
            page.is_title = True

        return page

class DescriptionScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        description = process_description_tag(soup_copy)
        if description:
            page.meta_description = description
            page.is_description = True

        return page


from page_scraper.core.node_utils import has_type, has_types
from page_scraper.entities.models import Entity, PageContext
from page_scraper.entities.builders import build_entity


class ArticleDetector:
    ARTICLE_TYPES = (
        "Article",
        "NewsArticle",
        "TechArticle",
        "ScholarlyArticle",
        "BlogPosting",
        "Report",
        "Blog"
    )

    def detect(self, page: PageContext) -> list[Entity]:
        articles = []
        for entry in page.nodes:
            node = entry['node']
            if has_types(node,*self.ARTICLE_TYPES) or (not articles and has_type(node, "WebPage") and node.get("url")):
                articles.append(build_entity(node, "Article"))

        return articles


from typing import Protocol
from page_scraper.entities.models import PageContext
from page_scraper.infrastructure.fetcher import HttpFetcher


class ScraperStep(Protocol):
    def run(self, page: PageContext) -> PageContext:
        ...


from .content.heading import HeadingScraper
from .content.image import ImageScraper
from .content.video import VideoTagScraper, VideoIframeScraper, VideoEmbedScraper
from .content.audio import AudioTagScraper
from .core.node_utils import get_categories, extract_breadcrumbs
from .core.pipeline import *
from .detectors.article import ArticleDetector
from .detectors.category import CategoryPageDetector
from .detectors.contact import ContactPageDetector
from .detectors.domain import DomainDetector
from .detectors.navigation import NavDetector
from .detectors.organization import OrgDetector
from .detectors.person import PersonDetector
from .detectors.product import ProductDetector
from .detectors.service import ServiceDetector
from .entities.builders import build_page_context
from .infrastructure.fetcher import HttpFetcher
from .infrastructure.parser import PageParser
from .meta.robots import RobotScraper
from .meta.social_tags import OgTitleScraper, OgDescriptionScraper, OgTypeScraper, OgSiteNameScraper, ModifiedTimeScraper
from .meta.title_meta import TitleScraper, DescriptionScraper
from .links.utils import prepare_response_header, get_clean_link
from .links.navigation import GetNavLinks
from .links.internal import InternalScraper
from .links.canonical import CanonicalScraper
from .version import __version__


from page_scraper.core.node_utils import has_type, path_contains, has_attrs
from page_scraper.entities.models import Entity, PageContext
from page_scraper.entities.builders import build_entity

class PersonDetector:
    def detect(self, page: PageContext) -> list[Entity]:
        persons = []
        seen = set()

        for entry  in page.nodes:
            node = entry['node']
            path = entry['path']

            if not has_type(node, "Person"):
                continue

            if path_contains(path, "Review", "review", "aggregateRating", "offers", "reviewRating", "Person", "person"):
                continue

            if not has_attrs(node,"url","@id","name"):
                continue

            entity = build_entity(node, "Person")

            key = (entity.url, (entity.name or "").lower())

            if key in seen:
                continue

            seen.add(key)
            persons.append(entity)

        return persons


from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from bs4 import BeautifulSoup
from requests.structures import CaseInsensitiveDict


@dataclass
class Entity:
    id: str
    name: str | None
    type: str
    alternate_name: str | None
    url: str | None
    description: str | None
    raw: dict
    date_published: datetime | None = None
    date_modified: datetime | None = None

@dataclass
class ResponseMeta:
    content_length: Optional[int]
    content_encoding: Optional[str]
    cache_control: Optional[str]
    last_modified: Optional[str]
    etag: Optional[str]
    server: Optional[str]


@dataclass
class UrlContext:
    url: str
    text: str | None = None
    status_code: int | None = None
    content_type: str | None = None
    position: str | None = None

@dataclass
class DomainContext:
    url: str
    ssl: bool = field(init=False, default=False)
    brand_name: str = field(init=False, default=None)
    brand_description: str = field(init=False, default=None)
    connections: list[str] = field(init=False, default=None)
    entities: list[Entity] = field(init=False, default=None)

@dataclass
class PageHeading:
    tag: str | None = None
    content: str | None = None

@dataclass
class PageMetaContext:
    name: str | None = None
    meta_type: str | None = None
    value: str | None = None

@dataclass
class PageContext:
    url: str
    soup: BeautifulSoup | None = None
    canonical_url: str | None = None
    nodes: list[dict] | None = None # all JSON-LD blocks parsed
    index_nodes: dict[str, dict] | None = None
    headers: CaseInsensitiveDict | None = None
    breadcrumbs: list | None = None
    # links: list[dict] | None = None
    meta_title: str | None = None
    is_title: bool = False
    meta_description: str | None = None
    is_description: bool = False
    is_index: bool = False
    is_follow: bool = False
    is_file: bool = False
    is_navigation: bool = False
    is_no_h1: bool = False
    is_domain: bool = False
    extera_h1: bool = False
    images_count: int = 0
    videos_count: int = 0
    audio_count: int = 0
    status_code: int = 0
    links: list[UrlContext] | None = None
    entities: list[Entity] = field(init=False, default=None)
    page_metas: list[PageMetaContext] = field(init=False, default=None)
    author: Entity | None = field(init=False, default=None)
    parent: dict | None = field(init=False, default=None)
    response_meta: ResponseMeta | None = field(init=False, default=None)
    headings : list[PageHeading] | None = field(init=False, default=None)



from page_scraper.core.node_utils import has_type
from page_scraper.entities.models import Entity, PageContext
from page_scraper.entities.builders import build_entity

class ServiceDetector:
    def detect(self, page: PageContext) -> list[Entity]:
        services = []
        for entry  in page.nodes:
            node = entry['node']

            if not has_type(node,"Service"):
                continue
            services.append(build_entity(node, "Service"))

        return services


from typing import Protocol
from page_scraper.entities.models import Entity
from page_scraper.entities.models import PageContext


class EntityDetector(Protocol):
    def detect(self, page) -> list[Entity]:
        ...

class PageStep(Protocol):
    def run(self,page: "PageContext") -> PageContext:
        ...


import copy

from page_scraper.entities.models import PageContext
from page_scraper.infrastructure.fetcher import HttpFetcher
from page_scraper.links.utils import get_clean_link, prepare_links
from page_scraper.core.utils import canonical_domain


class InternalScraper:
    def run(self,page:PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)
        clean_links = get_clean_link(soup_copy)

        domain = canonical_domain(page.url)
        result = prepare_links(domain, clean_links,"internal")
        page.links.extend(result)

        return page




from bs4 import Tag, BeautifulSoup


def get_meta_by_name(soup:BeautifulSoup, name:str):
    return soup.find_all('meta', attrs={'name': name})

def process_meta_text(tag:Tag):
    content = tag.get('content')
    if not content:
        return None
    return content.strip().replace("\n", " ")

def process_title_tag(soup:BeautifulSoup) -> str | None:
    title = soup.find('title')
    if not title:
        return None
    return process_meta_text(title)

def process_description_tag(soup:BeautifulSoup) -> str | None:
    description = soup.find('meta', attrs={'name': 'description'})
    if not description:
        return None
    return process_meta_text(description)

def process_property_tags(soup:BeautifulSoup, property_name:str) -> str | None:
    tag = soup.find('meta', attrs={'property': property_name})
    if not tag:
        return None
    return process_meta_text(tag)


# def resolve_author(article_node: dict, entities: dict):
#     author = article_node.get("author")
#
#     if isinstance(author, dict) and "@id" in author:
#         return entities.get(author["@id"])
#
#     if isinstance(author, dict):
#         return author
#
#     return None

def resolve_author(article_node: dict, entities_by_id: dict[str, dict]) -> dict | None:
    author = article_node.get("author")
    if not author:
        return None

    # Case 1 — author is a reference by @id
    if isinstance(author, dict) and "@id" in author:
        resolved = entities_by_id.get(author["@id"])
        if resolved:
            return resolved

        # Fallback: return minimal author info if Person not found
        return {
            "@type": "Person",
            "@id": author["@id"],
            "name": author.get("name"),
            "url": author.get("url"),
        }

    # Case 2 — author is inline object
    if isinstance(author, dict):
        return author

    # Case 3 — author is a string (sometimes happens)
    if isinstance(author, str):
        return {
            "@type": "Person",
            "name": author
        }

    # Case 4 - author is list
    if isinstance(author, list):
        return {
            "@type": "Person",
            "name": author[0].get("name"),
            "url": author[0].get("url"),
        }

    return None



from bs4 import BeautifulSoup
import json

from page_scraper.core.node_extractor import extract_nodes
from page_scraper.core.node_utils import build_entity_index
from page_scraper.core.utils import is_root_url


class PageParser:
    def parse(self, response):
        if response is None:
            return None

        soup = BeautifulSoup(response.content, "html.parser")
        is_domain = is_root_url(response.url)
        nodes = self.extract_jsonld(soup)
        return {
            "url": response.url,
            "status_code": response.status_code,
            "headers": response.headers,
            "nodes": nodes,
            "index_nodes": build_entity_index(nodes),
            "is_domain": is_domain,
        }

    def extract_jsonld(self, soup:BeautifulSoup):
        blocks = []
        for block in soup.find_all("script", type="application/ld+json"):
            try:
                blocks.append(json.loads(block.string))
            except Exception as e:
                pass

        nodes = extract_nodes(blocks)

        return nodes


import copy

from page_scraper.entities.builders import build_page_meta
from page_scraper.entities.models import PageContext
from page_scraper.meta.utils import process_property_tags


class OgTitleScraper:
    def run(self, page: PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        title = process_property_tags(soup_copy, 'og:title')
        if title and not page.meta_title:
            page.meta_title = title

        return page


class OgDescriptionScraper:
    def run(self, page: PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        description = process_property_tags(soup_copy, 'og:description')
        if description and not page.meta_description:
            page.meta_description = description

        return page

class OgTypeScraper:
    def run(self, page: PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        page_type = process_property_tags(soup_copy, 'og:type')
        if page_type:
            meta_context = build_page_meta('social','type', page_type)
            page.page_metas.append(meta_context)

        return page

class OgSiteNameScraper:
    def run(self, page: PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        page_type = process_property_tags(soup_copy, 'og:site_name')
        if page_type:
            meta_context = build_page_meta('social','site_name', page_type)
            page.page_metas.append(meta_context)

        return page

class ModifiedTimeScraper:
    def run(self, page: PageContext) -> PageContext:
        soup_copy = copy.copy(page.soup)

        modified_time = process_property_tags(soup_copy, 'article:modified_time')
        if modified_time:
            page.modified_time = modified_time

        return page